local prefabs = 
{
    "log",
	"charcoal",
	"robot_leaf_fx",
	"green_leaves_chop",
	"cutgrass",
	"twigs",
	"foliage",
	"rope",
}

local assets =
{
    Asset("ANIM", "anim/pillar_tree.zip"),
    Asset("SOUND", "sound/tentacle.fsb"),
    Asset("MINIMAP_IMAGE", "pillar_tree"),
	Asset("SCRIPT", "scripts/prefabs/canopyshadows.lua"),
}

local MIN = TUNING.SHADE_CANOPY_RANGE_SMALL
local MAX = MIN + TUNING.TREE_PILLAR_CANOPY_BUFFER
local NEW_VINES_SPAWN_RADIUS_MIN = 6
local DROPPED_ITEMS_SPAWN_HEIGHT = 10
local DROP_ITEMS_DIST_MIN = 6
local DROP_ITEMS_DIST_VARIANCE = 10

local function OnFar(inst)
    GetWorld():PushEvent("onchangecanopyzone", {})	
end

local function OnNear(inst)
    GetWorld():PushEvent("onchangecanopyzone", {})
end

local function DropLog(inst)
    if inst.logs > 0 then     
        local x, y, z = inst.Transform:GetWorldPosition()
        local item = SpawnPrefab("log")
        local dist = 0 + (math.random()*6)
        local theta = math.random() * 2 * PI
        local spawn_x, spawn_z
        spawn_x, spawn_z = x + math.cos(theta) * dist, z + math.sin(theta) * dist
        item.Transform:SetPosition(spawn_x, DROPPED_ITEMS_SPAWN_HEIGHT, spawn_z)    
        inst.logs = inst.logs - 1
    end
end

local function DropLogs(inst)
    DropLog(inst)
    DropLog(inst)
    if inst.logs < 1 then
        inst.logs = nil
        inst.drop_logs_task:Cancel()
        inst.drop_logs_task = nil        
    else
        inst.drop_logs_task = inst:DoTaskInTime(0.05, function() DropLogs(inst) end)
    end    
end

local function DropCanopyResources(inst)

    local x, _, z = inst.Transform:GetWorldPosition()
	
    local canopy_resources = {"cutgrass", "twigs", "foliage"}
    local item = SpawnPrefab(GetRandomItem(canopy_resources))
	
    local dist = DROP_ITEMS_DIST_MIN + DROP_ITEMS_DIST_VARIANCE * math.random()
    local theta = math.random() * 2 * PI

    local spawn_x, spawn_z

    spawn_x, spawn_z = x + math.cos(theta) * dist, z + math.sin(theta) * dist
	
    item.Transform:SetPosition(spawn_x, DROPPED_ITEMS_SPAWN_HEIGHT, spawn_z)
	item:DoTaskInTime(1, function()
	    if not item:IsOnValidGround() then
		    local x, y, z = item.Transform:GetWorldPosition()
		    SpawnPrefab("splash_clouds_drop").Transform:SetPosition(x, y, z)
		    item:Remove()			    
	    elseif item:GetIsOnWater() then
	        if item.components.floatable ~= nil then
		        item.components.floatable:OnHitWater()
		    else
		        local x, y, z = item.Transform:GetWorldPosition()
		        SpawnPrefab("splash_water_sink").Transform:SetPosition(x, y, z)
			    item:Remove()
		    end
	    end
	end)
	
end

local function chop_tree(inst, chopper, chopsleft)
	if chopper and chopper.components.beaverness and chopper.components.beaverness:IsBeaver() then
		inst.SoundEmitter:PlaySound("dontstarve/characters/woodie/beaver_chop_tree")
	else
		inst.SoundEmitter:PlaySound("dontstarve/wilson/use_axe_tree")
	end

	local player = GetClosestInstWithTag("player", inst, 40)
    if player then    
        player.components.playercontroller:ShakeCamera(inst, "VERTICAL", 0.25, 0.03, 0.2, 40)
    end

	local fx = math.random()<0.5 and SpawnPrefab("green_leaves_chop") or SpawnPrefab("robot_leaf_fx")
	local x, y, z = inst.Transform:GetWorldPosition()
	fx.Transform:SetPosition(x, y + 2 + math.random()*8, z)
end

local VINES_MUST = {"hangingvine"}
local function OnRemoveEntity(inst)
    local point = inst:GetPosition()
    local vines = TheSim:FindEntities(point.x, point.y, point.z, MAX+1, VINES_MUST)
    if #vines > 0 then
        for i, ent in ipairs(vines) do
            ent.AnimState:PlayAnimation("up")
			ent.SoundEmitter:PlaySound("dontstarve/cave/rope_up")
			if not ent.components.lootdropper then
			    ent:AddComponent("lootdropper")
			    ent.components.lootdropper:SetLoot({"rope"})
			end
			ent.components.lootdropper:DropLoot(ent:GetPosition())
			ent:DoTaskInTime(1, ent.Remove)
        end
    end
end

local function chop_down_tree(inst, chopper)
    if inst.droppedvines == nil then
        inst.droppedvines = true
    end	
    inst.components.lootdropper:DropLoot(inst:GetPosition())
    inst.logs = 15
    DropLogs(inst)
    inst:DoTaskInTime(0.5, function() 
	    local player = GetClosestInstWithTag("player", inst, 40)
	    if player then    
            player.components.playercontroller:ShakeCamera(inst, "VERTICAL", 0.25, 0.03, 0.6, 40)
        end	
	end)	
	for i=1, math.random(8,12) do
	    local fx = math.random()<0.5 and SpawnPrefab("green_leaves_chop") or SpawnPrefab("robot_leaf_fx")
	    local x, y, z = inst.Transform:GetWorldPosition()
	    fx.Transform:SetPosition(x, y + 2 + math.random()*8, z)
	end
	inst:DoTaskInTime(0.25, function() inst:Remove() end)
end

local function spawnvine(inst)
    local x, _, z = inst.Transform:GetWorldPosition()
    local radius_variance = MAX - NEW_VINES_SPAWN_RADIUS_MIN
    local vine = SpawnPrefab("hanging_vine")
    local theta = math.random() * PI * 2
    local offset = NEW_VINES_SPAWN_RADIUS_MIN + radius_variance * math.random()
    vine.Transform:SetPosition(x + math.cos(theta) * offset, 0, z + math.sin(theta) * offset)
end

local VINE_TAGS = { "hangingvine" }
local function SpawnMissingVines(inst)
    local x, _, z = inst.Transform:GetWorldPosition()
    local ents = TheSim:FindEntities(x, 0, z, MAX, VINE_TAGS)
    local num_existing_vines = ents ~= nil and #ents or 0

    if num_existing_vines < 1  then 

        local num_new_vines = math.random(1,2)

        for i=1,num_new_vines do
            spawnvine(inst)
        end
    end
end

local function CanTreePillarCollideWithOthers(inst, other)
	if other == nil 
	    or (not other:IsValid()) 
		or (not other.entity:IsVisible()) 
		or other:HasTag("INTERIOR_LIMBO")
		or other:HasTag("canopymob")
		or other.components.inventoryitem ~= nil
		or (other:IsInLimbo()) then
        return false
	end
    if other ~= GetPlayer() then
	    if other.components.locomotor ~= nil then
		    if not other.components.locomotor.disable then
			    return true
			end
		end
	else
	    local curr_speed = nil
	    if GetPlayer().components.locomotor ~= nil then
		    curr_speed = GetPlayer().components.locomotor:GetRunSpeed()
		end
		if curr_speed == nil then
		    return false
		elseif curr_speed >= TUNING.TREE_PILLAR_COLLIDE_REQUIRED_PLAYER_SPEED then -- 12
		    return true
		end
	end
	return false
end

local function OnCollide(inst, other)
    if not inst:HasTag("cooldown_on_collide") and CanTreePillarCollideWithOthers(inst, other) then
        if other and other == GetPlayer() then
	        local x, y, z = inst.Transform:GetWorldPosition()
		    local CanopyEnts = TheSim:FindEntities(x,y,z, MIN, {"canopymob"})	
	        for k,v in pairs(CanopyEnts) do
			    if v.components.combat ~= nil then
				    v.components.combat:SuggestTarget(GetPlayer())
				end
			end
	    end
	    local player = GetClosestInstWithTag("player", inst, 40)
        if player then    
            player.components.playercontroller:ShakeCamera(inst, "VERTICAL", 2.8, 0.025, 0.3, 40)
        end
        inst:DoTaskInTime(0.65, SpawnMissingVines)
		for i=1, math.random(3,5) do
		    DropCanopyResources(inst)
		end
	    inst:AddTag("cooldown_on_collide")
	    inst.components.timer:StartTimer("cooldown_on_collide", TUNING.FREQUENCY)
        inst:ListenForEvent("timerdone", function(inst, data) 
		    if inst:HasTag("cooldown_on_collide") then
			    inst:RemoveTag("cooldown_on_collide")
		    end
	    end)		
	end
end

local function OnBurnt(inst)
    inst.components.lootdropper:SetLoot({"charcoal", "charcoal", "charcoal", "charcoal", "charcoal", "charcoal", "charcoal"})
    inst.components.lootdropper:DropLoot(inst:GetPosition())
    local player = GetClosestInstWithTag("player", inst, 40)
    if player then    
        player.components.playercontroller:ShakeCamera(inst, "VERTICAL", 0.25, 0.003, 0.6, 40)
    end
	for i=1, math.random(8,12) do
	    local fx = math.random()<0.5 and SpawnPrefab("green_leaves_chop") or SpawnPrefab("robot_leaf_fx")
	    local x, y, z = inst.Transform:GetWorldPosition()
	    fx.Transform:SetPosition(x, y + 2 + math.random()*8, z)
	end	
	inst:DoTaskInTime(0.1, function() inst:Remove() end)
end

local function startvines(inst)
    inst:DoTaskInTime(2 + math.random(),function() spawnvine(inst) end)
    inst:DoTaskInTime(2.5 + math.random(),function() spawnvine(inst) end)
    inst:DoTaskInTime(2.5 + math.random(),function() spawnvine(inst) end)
    if math.random() < 0.33 then 
        inst:DoTaskInTime(3 + math.random(),function() spawnvine(inst) end)
    end
    inst.droppedvines = true
end

local function OnSave(inst, data)	
	if inst:HasTag("cooldown_on_collide") then
	    print("Save tag of cooldown_on_collide")
	    data.cooldown_on_collide = true
	end
	
	if inst.components.timer ~= nil then
	    if inst.components.timer:TimerExists("cooldown_on_collide") then
		    print("Save artificial tree pillar timer time left")
		    data.timer_cooldown_on_collide = inst.components.timer:GetTimeLeft("cooldown_on_collide")
		end
	end	

    if inst.droppedvines then
        data.droppedvines = true
    end

end

local function OnLoad(inst, data)	
	if data ~= nil and data.cooldown_on_collide then
	    if not inst:HasTag("cooldown_on_collide") then
		    print("Add tag of cooldown_on_collide")
		    inst:AddTag("cooldown_on_collide")
		end
	end	
	
	if data ~= nil and data.timer_cooldown_on_collide then
	    if inst.components.timer ~= nil then
		    if inst.components.timer:TimerExists("cooldown_on_collide") then
		        print("Load artificial tree pillar timer time")
				inst.components.timer:SetTimeLeft("cooldown_on_collide", data.timer_cooldown_on_collide)
		        inst:ListenForEvent("timerdone", function(inst, data) 
			        if inst:HasTag("cooldown_on_collide") then
			            inst:RemoveTag("cooldown_on_collide")
			        end
		        end)				
		    else
			    print("Load missing artificial tree pillar timer time")
			    inst.components.timer:StartTimer("cooldown_on_collide", data.timer_cooldown_on_collide)
		        inst:ListenForEvent("timerdone", function(inst, data) 
			        if inst:HasTag("cooldown_on_collide") then
			            inst:RemoveTag("cooldown_on_collide")
			        end
		        end)
		    end
		end
	end	
	
    if data ~= nil and data.droppedvines then
        inst.droppedvines = true
    end

end

local function fn(Sim)
	local inst = CreateEntity()
	local trans = inst.entity:AddTransform()
	local anim = inst.entity:AddAnimState()
    inst.entity:AddSoundEmitter()

    MakeObstaclePhysics(inst, 1.5, 12)   --3, 24

    trans:SetScale(0.9,1,0.9)   --1,1,1

    inst:AddTag("plant")	
    inst:AddTag("shadecanopysmall")
	inst:AddTag("Canopy")
	
	local minimap = inst.entity:AddMiniMapEntity()
	minimap:SetIcon( "pillar_tree.png" )

	anim:SetBank("pillar_tree")
	anim:SetBuild("pillar_tree")   

    anim:PlayAnimation("idle",true)
	
	inst.Physics:SetCollisionCallback(OnCollide)
	
    inst:AddComponent("inspectable")

    inst:AddComponent("timer")

    inst:AddComponent("mystery")

    inst:AddComponent("lootdropper")
    inst.components.lootdropper:SetLoot({"log", "log", "log", "log", "twigs", "twigs", "twigs"})

    inst:AddComponent("workable")
    inst.components.workable:SetWorkAction(ACTIONS.CHOP)
    inst.components.workable:SetMaxWork(TUNING.TREE_PILLAR_CHOPS)
    inst.components.workable:SetWorkLeft(TUNING.TREE_PILLAR_CHOPS)
    inst.components.workable:SetOnWorkCallback(chop_tree)
    inst.components.workable:SetOnFinishCallback(chop_down_tree)
 
    inst:AddComponent("distancefade")
    inst.components.distancefade:Setup(15,25)

--[[ FROM DST. DOES NOT WORK WELL WITH DSV.
    inst:AddComponent("canopyshadows")
    inst.components.canopyshadows.range = math.floor(TUNING.SHADE_CANOPY_RANGE_SMALL/4)
]]

    inst:AddComponent("playerprox")
    inst.components.playerprox:SetDist(MIN, MAX)
    inst.components.playerprox:SetOnPlayerFar(OnFar)
    inst.components.playerprox:SetOnPlayerNear(OnNear)
	
    MakeLargeBurnable(inst)
    MakeMediumPropagator(inst)
    inst.components.burnable:SetFXLevel(6)
    inst.components.burnable:SetOnBurntFn(OnBurnt)
	
	inst:AddComponent("shadowcaster")
	inst.components.shadowcaster:setrange(22)

    inst:DoTaskInTime(0.5, function() 
		if not inst.droppedvines then   
		    startvines(inst) 
		end
	end)

	inst.OnSave = OnSave
    inst.OnLoad = OnLoad

    inst:ListenForEvent("onremove", OnRemoveEntity)

    return inst
end

return Prefab("tree_pillar_artificial", fn, assets, prefabs)
