local assets =
{
    Asset("ANIM", "anim/oceantreenut.zip"),
	--Asset("MINIMAP_IMAGE", "bramble"),
	--Asset("INV_IMAGE", "burr"),
}

local prefabs =
{
	"twigs",
	"boneshard",
}

SetSharedLootTable('treenut',
{
    {'twigs',	      1.00},
    {'twigs',		  0.50},
    {'boneshard',     1.00},
})

local PHYSICS_RADIUS = .75
local ATTEMPT_GROW_FREQUENCY = 60

local function OnWorkedFinished(inst, worker)
    inst.components.lootdropper:DropLoot()

    local fx = SpawnPrefab("collapse_small")
    fx.Transform:SetPosition(inst.Transform:GetWorldPosition())
	
    inst.SoundEmitter:PlaySound("dontstarve/common/destroy_stone")	

    inst:Remove()
end

local function OnUnequip(inst, owner)
    owner.AnimState:ClearOverrideSymbol("swap_body")
end

local function OnEquip(inst, owner)
    owner.AnimState:OverrideSymbol("swap_body", "oceantreenut", "swap_body")
end

local function CanGrow(inst)
    if inst.components.inventoryitem ~= nil and inst.components.inventoryitem.owner ~= nil then return false end
	if inst:IsNear(GetPlayer(), 40) then return false end
    local pt = inst:GetPosition()
	local ground_OK = inst:GetIsOnLand(pt.x, pt.y, pt.z)
	local tiletype = GetGroundTypeAtPosition(pt)
	if ground_OK and (tiletype == GROUND.DEEPRAINFOREST or tiletype == GROUND.DEEPRAINFOREST_NOCANOPY) then
	    local pos = Vector3(inst.Transform:GetWorldPosition())
	    local ents = TheSim:FindEntities(pos.x,pos.y,pos.z, 16, nil, {'NOBLOCK', 'player', 'FX'}) 
	    for k, v in pairs(ents) do
			if v ~= inst and v:IsValid() and v.entity:IsVisible() and not v.components.placer and v.parent == nil then
				if distsq(Vector3(v.Transform:GetWorldPosition()), pt) < 4*4 then
					return false
				end		
			end
		end
		return true
	end
	return false
end

local function Grow(inst)
    print("grow into tree_pillar_artificial")
	if GetPlayer() and GetPlayer():HasTag("plantkin") then
        if GetPlayer().growplantfn then
            GetPlayer().growplantfn(GetPlayer())
        end	
	end
    local tree = SpawnPrefab("tree_pillar_artificial")  
    local x, y, z = inst.Transform:GetWorldPosition()
    tree.Transform:SetPosition(x, y, z)
    inst:Remove()
end

local function AttemptGrow(inst)
    if CanGrow(inst) then
	    print("ready to grow")
        Grow(inst)
    end
end

local function OnTimerDone(inst, data)
    if CanGrow(inst) then
	    print("ready to grow")
        Grow(inst)
    else
	    print("fail to grow, check grow periodically")
	    inst.should_attempt_grow = true
	    inst.attempt_grow_task = inst:DoPeriodicTask(ATTEMPT_GROW_FREQUENCY, AttemptGrow)
    end
end

local function OnSave(inst, data)
    if inst.should_attempt_grow then
	    print("Save attempt grow")
        data.should_attempt_grow = true
    end
	if inst.components.timer ~= nil then
	    if inst.components.timer:TimerExists("grow") then
		    print("Save tree nut timer time left")
		    data.timer_grow = inst.components.timer:GetTimeLeft("grow")
		end
	end
end

local function OnLoad(inst, data)
    if data ~= nil and data.should_attempt_grow then
	    print("Load attempt grow")
        inst.should_attempt_grow = true
		if inst.attempt_grow_task == nil then
		    print("Load attempt grow task")
		    inst.attempt_grow_task = inst:DoPeriodicTask(ATTEMPT_GROW_FREQUENCY, AttemptGrow)
		end
    end
	if data ~= nil and data.timer_grow then
	    if inst.components.timer ~= nil then
		    if inst.components.timer:TimerExists("grow") then
		        print("Load tree nut timer time")
				inst.components.timer:SetTimeLeft("grow", data.timer_grow)
		        inst:ListenForEvent("timerdone", OnTimerDone)				
		    else
			    print("Load missing tree nut timer time")
			    inst.components.timer:StartTimer("grow", data.timer_grow)
		        inst:ListenForEvent("timerdone", OnTimerDone)
		    end
		end
	end
end

local function MakeHeavyObstaclePhysics(inst, rad, height)
    inst:AddTag("blocker")
    inst.entity:AddPhysics()
    --inventory physics
    inst.Physics:SetFriction(.1)
    inst.Physics:SetDamping(0)
    inst.Physics:SetRestitution(0)
    --obstacle physics
    inst.Physics:SetMass(0)
    inst.Physics:SetCollisionGroup(COLLISION.OBSTACLES)
    inst.Physics:ClearCollisionMask()
	inst.Physics:CollidesWith(GetWorldCollision())
	inst.Physics:CollidesWith(GetWaterCollision())
    inst.Physics:CollidesWith(COLLISION.ITEMS)
    inst.Physics:CollidesWith(COLLISION.CHARACTERS)
	inst.Physics:CollidesWith(COLLISION.INTWALL)
    inst.Physics:SetCapsule(rad, height or 2)
end

local function RemoveSpeedMult(inst)
    if inst.components.equippable ~= nil then
        inst.components.equippable.walkspeedmult = nil
	end
end

local function AddSpeedMult(inst)
    if inst.components.equippable ~= nil then
        inst.components.equippable.walkspeedmult = TUNING.HEAVY_SPEED_MULT
	end
end

local function fn()
    local inst = CreateEntity()

    inst.entity:AddTransform()
    inst.entity:AddAnimState()
	inst.entity:AddSoundEmitter()
	
    --MakeInventoryPhysics(inst)
	MakeHeavyObstaclePhysics(inst, PHYSICS_RADIUS)
	inst.physicsradiusoverride = PHYSICS_RADIUS
	
    inst.AnimState:SetBank("oceantreenut")
    inst.AnimState:SetBuild("oceantreenut")
    inst.AnimState:PlayAnimation("idle")

	local minimap = inst.entity:AddMiniMapEntity()
	minimap:SetIcon("tree_nut.tex")

    inst:AddTag("heavy")
	inst:AddTag("treenut")
	inst:AddTag("plant")

    inst.should_attempt_grow = false
    inst.attempt_grow_task = nil
	
    inst:AddComponent("inspectable")

    inst:AddComponent("locomotor")
    inst.components.locomotor.disable = true
	
	inst:AddComponent("lootdropper")
    inst.components.lootdropper:SetChanceLootTable('treenut')

    inst:AddComponent("inventoryitem")
    inst.components.inventoryitem.atlasname = "images/inventoryimages/tree_nut.xml"
	inst.components.inventoryitem.imagename = "tree_nut"
    inst.components.inventoryitem.cangoincontainer = false
	inst.components.inventoryitem.nobounce = true
	inst.components.inventoryitem:SetOnPutInInventoryFn(function(inst)
	    if inst.components.workable ~= nil then
		    inst.components.workable:SetWorkable(false)
		end
	end)
	inst.components.inventoryitem:SetOnDroppedFn(function(inst, dropper)
	    if inst.components.workable ~= nil then
		    inst.components.workable:SetWorkable(true)
		end
	end)

    inst:AddComponent("heavyobstaclephysics")
    inst.components.heavyobstaclephysics:SetRadius(PHYSICS_RADIUS)
	
    inst:AddComponent("equippable")
    inst.components.equippable.equipslot = EQUIPSLOTS.BODY
    inst.components.equippable:SetOnEquip(OnEquip)
    inst.components.equippable:SetOnUnequip(OnUnequip)
    inst.components.equippable.walkspeedmult = TUNING.HEAVY_SPEED_MULT
	
    inst:AddComponent("workable")
    inst.components.workable:SetWorkAction(ACTIONS.HAMMER)
    inst.components.workable:SetWorkLeft(1)
	inst.components.workable:SetOnFinishCallback(OnWorkedFinished)

    inst:AddComponent("timer")
	inst:DoTaskInTime(0.5, function()
	    if not inst.components.timer:TimerExists("grow") and not inst.should_attempt_grow then
            inst.components.timer:StartTimer("grow", TUNING.TREENUT_GROW_TIME + math.random() * TUNING.TREENUT_GROW_TIME_VARIANCE)
	    end
	end)
    inst:ListenForEvent("timerdone", OnTimerDone)

    inst:ListenForEvent("mounted", RemoveSpeedMult, GetPlayer())
    inst:ListenForEvent("dismounted", AddSpeedMult, GetPlayer())

    inst.OnSave = OnSave
    inst.OnLoad = OnLoad

    return inst
end

return Prefab("treenut", fn, assets, prefabs)
