local assets =
{
    Asset("ANIM", "anim/pollen.zip"),
	Asset("INV_IMAGE", "pollen"),
}

local prefabs =
{

}

local ATTEMPT_GROW_FREQUENCY = 30

local function CanGrow(inst)
    if inst.components.inventoryitem ~= nil and inst.components.inventoryitem.owner ~= nil then return false end
	if inst:IsNear(GetPlayer(), 40) then return false end
	if inst:HasTag("fire") then return false end
    local pt = inst:GetPosition()
	local ground_OK = inst:GetIsOnLand(pt.x, pt.y, pt.z)
	local tiletype = GetGroundTypeAtPosition(pt)
	if ground_OK and tiletype == GROUND.PAINTED then
	    local pos = Vector3(inst.Transform:GetWorldPosition())
	    local ents = TheSim:FindEntities(pos.x,pos.y,pos.z, 16, nil, {'NOBLOCK', 'player', 'FX'}) 
	    for k, v in pairs(ents) do
			if v ~= inst and v:IsValid() and v.entity:IsVisible() and not v.components.placer and v.parent == nil then
				if distsq(Vector3(v.Transform:GetWorldPosition()), pt) < 4*4 then
					return false
				end		
			end
		end
		return true
	end
	return false
end

local function Grow(inst)
    print("grow into a tuber tree")
	if GetPlayer() and GetPlayer():HasTag("plantkin") then
        if GetPlayer().growplantfn then
            GetPlayer().growplantfn(GetPlayer())
        end	
	end
    local tree = SpawnPrefab("tubertree_short")  
    local x, y, z = inst.Transform:GetWorldPosition()
    tree.Transform:SetPosition(x, y, z)
    inst:Remove()
end

local function AttemptGrow(inst)
    if CanGrow(inst) then
	    print("ready to grow")
        Grow(inst)
    end
end

local function OnTimerDone(inst, data)
    if CanGrow(inst) then
	    print("ready to grow")
        Grow(inst)
    else
	    print("fail to grow, check grow periodically")
	    inst.should_attempt_grow = true
	    inst.attempt_grow_task = inst:DoPeriodicTask(ATTEMPT_GROW_FREQUENCY, AttemptGrow)
    end
end

local function OnEaten_Raw(inst, eater)
    if eater.components.poisonable ~= nil and eater:HasTag("poisonable") then
        eater.components.poisonable:Poison()
    end 
	if eater.components.hayfever ~= nil and not eater:HasTag("plantkin") --[[and eater.components.hayfever.enabled]] then  
        if eater.sg then
			eater:DoTaskInTime(0.01, function() eater.sg:GoToState("sneeze") end)
			eater:DoTaskInTime(0.25, function() eater.sg:GoToState("sneeze") end)
			eater:DoTaskInTime(0.50, function() eater.sg:GoToState("sneeze") end)
		end
	end
end

local function OnEaten_Cooked(inst, eater)
    --if eater.components.poisonable ~= nil and eater:HasTag("poisonable") then
    --    eater.components.poisonable:Poison()
    --end 
	if eater.components.hayfever ~= nil and not eater:HasTag("plantkin") --[[and eater.components.hayfever.enabled]] then   
        if eater.sg then
			eater:DoTaskInTime(0.01, function() eater.sg:GoToState("sneeze") end)
			eater:DoTaskInTime(0.25, function() eater.sg:GoToState("sneeze") end)
			eater:DoTaskInTime(0.50, function() eater.sg:GoToState("sneeze") end)
		end
	end
end

local function OnSave(inst, data)
    if inst.should_attempt_grow then
	    print("Save attempt to grow")
        data.should_attempt_grow = true
    end
	if inst.components.timer ~= nil then
	    if inst.components.timer:TimerExists("grow") then
		    print("Save tuber seed timer time left")
		    data.timer_grow = inst.components.timer:GetTimeLeft("grow")
		end
	end	
end

local function OnLoad(inst, data)
    if data ~= nil and data.should_attempt_grow then
	    print("Load attempt to grow")
        inst.should_attempt_grow = true
		if inst.attempt_grow_task == nil then
		    print("Load attempt to grow task")
		    inst.attempt_grow_task = inst:DoPeriodicTask(ATTEMPT_GROW_FREQUENCY, AttemptGrow)
		end
    end
	if data ~= nil and data.timer_grow then
	    if inst.components.timer ~= nil then
		    if inst.components.timer:TimerExists("grow") then
		        print("Load tuber seed timer time")
				inst.components.timer:SetTimeLeft("grow", data.timer_grow)
		        inst:ListenForEvent("timerdone", OnTimerDone)				
		    else
			    print("Load missing tuber seed timer time")
			    inst.components.timer:StartTimer("grow", data.timer_grow)
		        inst:ListenForEvent("timerdone", OnTimerDone)
		    end
		end
	end
end

local function fn()
    local inst = CreateEntity()

    inst.entity:AddTransform()
    inst.entity:AddAnimState()
	inst.entity:AddSoundEmitter()
	
    inst.AnimState:SetBank("pollen")
    inst.AnimState:SetBuild("pollen")
    inst.AnimState:PlayAnimation("idle")

    MakeInventoryPhysics(inst)
    MakeInventoryFloatable(inst, "idle_water", "idle")
    MakeBlowInHurricane(inst, TUNING.WINDBLOWN_SCALE_MIN.LIGHT, TUNING.WINDBLOWN_SCALE_MAX.LIGHT)	

    inst.should_attempt_grow = false
    inst.attempt_grow_task = nil
	
    inst:AddTag("plant")
	inst:AddTag("poisonous")
    inst:AddTag("cattoy")
    inst:AddTag("tuber_tree_seed")	
	
    inst:AddComponent("tradable")
	
    inst:AddComponent("inspectable")

    inst:AddComponent("inventoryitem")
	inst.components.inventoryitem.imagename = "pollen"

    inst:AddComponent("stackable")
	inst.components.stackable.maxsize = TUNING.STACK_SIZE_SMALLITEM

    inst:AddComponent("cookable")
    inst.components.cookable.product = "tuber_pollen_cooked"
	
    inst:AddComponent("edible")
    inst.components.edible.hungervalue = TUNING.CALORIES_TINY
    inst.components.edible.healthvalue = TUNING.HEALING_TINY
    inst.components.edible.foodtype = "SEEDS"
    inst.components.edible.foodstate = "RAW"
	inst.components.edible:SetOnEatenFn(OnEaten_Raw)

    inst:AddComponent("bait")

    inst:AddComponent("timer")
	inst:DoTaskInTime(0.5, function()
	    if not inst.components.timer:TimerExists("grow") and not inst.should_attempt_grow then
            inst.components.timer:StartTimer("grow", TUNING.TUBERSEED_GROW_TIME + math.random() * TUNING.TUBERSEED_GROW_TIME_VARIANCE)
	    end
	end)
    inst:ListenForEvent("timerdone", OnTimerDone)

    MakeSmallBurnable(inst, TUNING.SMALL_BURNTIME)
    MakeSmallPropagator(inst)
	
    inst.OnSave = OnSave
    inst.OnLoad = OnLoad

    return inst
end

local function cooked()
    local inst = CreateEntity()
	
    inst.entity:AddTransform()
    inst.entity:AddAnimState()
    inst.entity:AddSoundEmitter()

    inst.AnimState:SetBank("pollen")
    inst.AnimState:SetBuild("pollen")
    inst.AnimState:PlayAnimation("cooked")
	
    MakeInventoryPhysics(inst)
    MakeInventoryFloatable(inst, "idle_cooked_water", "cooked")
	MakeBlowInHurricane(inst, TUNING.WINDBLOWN_SCALE_MIN.MEDIUM, TUNING.WINDBLOWN_SCALE_MAX.MEDIUM)
	
--[[
    inst.should_attempt_grow = false
    inst.attempt_grow_task = nil
]]	
    --inst:AddTag("plant")
	--inst:AddTag("poisonous")
    inst:AddTag("cattoy")
	inst:AddTag("tuber_tree_seed_cooked")
	
    inst:AddComponent("tradable")

    inst:AddComponent("inspectable") 
	
    inst:AddComponent("inventoryitem")
	inst.components.inventoryitem.imagename = "pollen"   --Not Cooked image

    inst:AddComponent("stackable")
	inst.components.stackable.maxsize = TUNING.STACK_SIZE_SMALLITEM

--[[
    inst:AddComponent("cookable")
    inst.components.cookable.product = "tuber_pollen_cooked"
]]
	
    inst:AddComponent("edible")
    inst.components.edible.hungervalue = TUNING.CALORIES_TINY
    inst.components.edible.healthvalue = TUNING.HEALING_SMALL
    inst.components.edible.foodtype = "SEEDS"
	inst.components.edible.foodstate = "COOKED"
	inst.components.edible:SetOnEatenFn(OnEaten_Cooked)

    inst:AddComponent("bait")

--[[
    inst:AddComponent("timer")
    inst.components.timer:StartTimer("grow", TUNING.TUBERSEED_GROW_TIME + math.random() * TUNING.TUBERSEED_GROW_TIME_VARIANCE)
    inst:ListenForEvent("timerdone", OnTimerDone)
]]
	
    MakeSmallBurnable(inst, TUNING.SMALL_BURNTIME)
    MakeSmallPropagator(inst)

--[[  
    inst.OnSave = OnSave
    inst.OnLoad = OnLoad
]] 
 
    return inst
end

return Prefab("tuber_pollen", fn, assets, prefabs),
       Prefab("tuber_pollen_cooked", cooked, assets)